﻿// ------------------------------------------------------------------------ //
//  TextDataToPaperChart
//      paperChart(NV.exe) へテキストデータを送信するプログラム
//  Ver 1.00  2014/08/18
//      by Tomohiko Saito (tomohiko.saito@gmail.com)
//  備考
//      NV.exeと外部プログラムのデータ通信は
//      Win32APIプロセス通信 WM_COPYDATA メッセージを使用
//      
//      外部プログラムは，NV.exe のウィンドウハンドルに対し，
//      WM_COPYDATA メッセージを送信する事で通信している
//
//      NV.exe は，dircnf.txt に記述された外部プログラムを起動する際
//      /std_arg/ パラメータが指定されると，NV.exeのウィンドウハンドルを
//      起動引数 hwnd=xxxx として 外部プログラムに通知する
//
//      本プログラムでは，dircnf.txt に 起動引数 /std_arg/ を指定した
//      起動方法の他，
// ------------------------------------------------------------------------ //
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

using System.Runtime.InteropServices;   // for Win32API
using System.Diagnostics;               // for Process

namespace TextDataToPaperChart
{
    public partial class FormMain : Form
    {
        //
        // WM_COPYDATA / SendMessage
        //
        const Int32 WM_COPYDATA = 0x4A;
        struct COPYDATASTRUCT
        {
            public IntPtr dwData;
            public UInt32 cbData;
            public IntPtr lpData;
        }

        [DllImport("user32.dll")]
        private static extern uint SendMessage(IntPtr hWnd, int msg, IntPtr wParam, ref COPYDATASTRUCT lParam);
        [DllImport("user32.dll")]
        private static extern uint SendMessage(IntPtr hWnd, int msg, IntPtr wParam, IntPtr lParam);
        [DllImport("kernel32.dll")]
        public static extern Int32 lstrlenA(IntPtr str);

        const Int32 WM_COMMAND = 0x111;

        //
        // マウスドラッグ操作で，直下のウィンドウを取得する WindowFromPoint
        //
        [DllImport("user32.dll")]
        public static extern IntPtr WindowFromPoint(Point pt);
        [DllImport("user32.dll")]
        public static extern IntPtr GetParent(IntPtr hWnd);

        delegate int EnumChildProc(IntPtr hwnd, IntPtr lParam);
        [DllImport("user32.dll")]
        private static extern int EnumChildWindows(IntPtr hWnd, EnumChildProc lpEnumFunc, IntPtr lParam);

        //
        // Mutex 操作
        //
        const Int32 MUTEX_ALL_ACCESS = 0x001F0001;

        [DllImport("kernel32.dll")]
        public static extern IntPtr OpenMutex(Int32 dwAccess, bool bInherit, string sName);
        [DllImport("kernel32.dll")]
        public static extern bool CloseHandle(IntPtr hHandle);

        // NV.exe が作成する Mutexの値
        const string MUTEX_MONITORING = "metaChartIsMonitoring";
        const string MUTEX_TOPMOSTWND = "metaChartHasTopmostWnd";
        const string MUTEX_TERPUMP = "terpumpIsRunning";
        const string MUTEX_OPENFILE = "openFileIsRunninng";
        const string MUTEX_MANUATTR = "manuAttrIsRunninng";
        const string MUTEX_QUICKDELIVERY = "quickDeliveryIsRunninng";

        //
        // 変数
        //
        IntPtr hNvHwnd = IntPtr.Zero;
        IntPtr hNvFrame = IntPtr.Zero;
        IntPtr hMyHwnd = IntPtr.Zero;

        //
        // Unix Time の処理
        //
        // unix epochをDateTimeで表した定数
        public readonly static DateTime UnixEpoch = new DateTime(1970, 1, 1, 0, 0, 0, DateTimeKind.Utc);

        // UNIX時間からDateTimeに変換するメソッド
        public static DateTime FromUnixTime(long unixTime)
        {
            // unix epochからunixTime秒だけ経過した時刻を求める
            return UnixEpoch.AddSeconds(unixTime);
        }
        public static DateTime FromUnixTime(string strTime)
        {
            // strTime を unixTime秒 に変換
            long unixTime = long.Parse(strTime);
            return UnixEpoch.AddSeconds(unixTime);
        }
        public static long GetUnixTime(DateTime strTime)
        {
            // UTC時間に変換
            strTime = strTime.ToUniversalTime();
            // UNIXエポックからの経過時間を取得
            TimeSpan elapsedTime = strTime - UnixEpoch;
            // 経過秒数に変換
            return (long)elapsedTime.TotalSeconds;
        }

        public FormMain()
        {
            InitializeComponent();
        }

        private void FormMain_Load(object sender, EventArgs e)
        {
            //コマンドラインを配列で取得する
            string[] args = System.Environment.GetCommandLineArgs();

            // 引数パラメータから，ウィンドウハンドルを取得
            foreach (string v in args)
            {
                const string _STR_HWND_ = "hwnd=";
                if (v.IndexOf(_STR_HWND_) == 0)
                {
                    // 与えられた引数の = 以降の文字列を16進文字列として変換し，
                    // NVのウィンドウハンドルを保存
                    hNvHwnd = (IntPtr)Convert.ToInt32(v.Substring(_STR_HWND_.Length), 16);
                    tbxNvHWnd.Text = hNvHwnd.ToString("X8");

                    hNvFrame = GetParent(hNvHwnd);
                    tbxFraHWnd.Text = (hNvFrame == IntPtr.Zero) ? "(null)" : hNvFrame.ToString("X8");
                }
            }

            // 自分自身のウィンドウハンドル
            hMyHwnd = this.Handle;

            // Idle時のイベントハンドラを設定する
            Application.Idle += new EventHandler(OnIdle);
            
            // 時刻用 DateTimePicker と UnixTime に，起動時刻をセット
            dtpEventTime.Value = DateTime.Now;
            tbxUnixTime.Text = GetUnixTime(DateTime.Now).ToString();

            //
            // コンボボックスへサンプルデータをセットする ... PpcComboItem
            //
            // 送信されるデータ形式は，paperChart が作成する xxxx.txt ファイルを参照のこと
            //
            // 数字部分は，イベント時刻を UnixTime に変換したもので，1970/01/01 0:00 からの経過秒数
            //
            // 以下のサンプルでは，/remarks/, /administration/ では，時刻指定を行うため
            // 書式制御文字列としている
            //
            PpcComboItem[] items = new PpcComboItem[] {            
                new PpcComboItem ("--- none ---", ""),
                new PpcComboItem ("demographs", "/demographs/\r\n"
                                                + "\tID==00001234\r\n"
                                                + "\t漢字姓名==平波 茶途\r\n"
                                                + "\t性別==女\r\n"
                                                + "\t生年月日==1990/01/02\r\n"
                                                + "\t身長==155\r\n"
                                                + "\t体重==48\r\n"
                                                + "\tABO型==O\r\n"
                                                + "\t術前病名1==慢性疲労症候群\r\n"
                                                + "\t実施術名1==睡眠による疲労除去術\r\n"),
                new PpcComboItem ("remarks (麻酔開始)", "/remarks/\r\n\t麻酔\r\n\t\t!\r\n\t\t()1b{0},麻酔が始まったよ\r\n"),
                new PpcComboItem ("remarks (麻酔終了)", "/remarks/\r\n\t麻酔\r\n\t\t()1e{0},\x11応答\x12+\x11呼吸\x12++\x11特記\x12オシマイ\r\n"),
                new PpcComboItem ("administrations (ﾌｪﾝﾀ)", "/administrations/\r\n\tﾌｪﾝﾀ,,,BPV,μg/hr,1\r\n\t\t(){0},2,*,100,2,B"),
                new PpcComboItem ("administrations (ｴｽﾗｯｸｽ)", "/administrations/\r\n\tｴｽﾗｯｸｽ,,,BP,mg,1\r\n\t\t(){0},3,*,50,5,B"),
                new PpcComboItem ("administrations (ｱﾙﾁﾊﾞ)", "/administrations/\r\n\tｱﾙﾁﾊﾞ,,,BP,μg/kg/min,0.01\r\n\t\t(){0},15,*,2000,20,P"),
            };

            // テキストサンプル コンボボックスへの値設定
            foreach (PpcComboItem item in items) cbxSelectSample.Items.Add(item);

            //
            // paperChartのメニューは，固定されたIDを持つメニュー項目と，
            // 設定ファイルからダイナミックに作成されるメニュー項目に分けられる
            // 以下は，NV.exe のメニューIDを解析して，基本的な項目を設定した
            //
            // paperChartメニュー コンボボックスへの値設定
            cbxPpcMenuItem.Items.Add(new PpcMenuItem("ﾓﾆﾀ開始(新規)", 32771));
            cbxPpcMenuItem.Items.Add(new PpcMenuItem("ﾓﾆﾀ再開(追記)", 32772));
            cbxPpcMenuItem.Items.Add(new PpcMenuItem("ﾓﾆﾀ停止(保存)", 32773));
            cbxPpcMenuItem.Items.Add(new PpcMenuItem("開く", 32774));
            cbxPpcMenuItem.Items.Add(new PpcMenuItem("上書き保存", 32775));
            cbxPpcMenuItem.Items.Add(new PpcMenuItem("終了", 57665));
        }

        //
        // マウス を Drag & Capture することで，マウス下のウィンドウを取得する
        //  → lblNvHWnd ラベルをドラッグすることで，キャプチャしウィンドウハンドルを取得する
        //
        private void lblNvHWnd_MouseDown(object sender, MouseEventArgs e)
        {
            // ラベルをクリックすると マウスキャプチャーを行う
            lblNvHWnd.Capture = true;
            Cursor.Current = Cursors.Hand;          // Cross;
        }

        private void lblNvHWnd_MouseMove(object sender, MouseEventArgs e)
        {
            if (!lblNvHWnd.Capture) return;

            // キャプチャー時は，マウスの下にあるウィンドウのハンドルを取得
            hNvHwnd = WindowFromPoint(Cursor.Position);
            tbxNvHWnd.Text = hNvHwnd.ToString("X8");
        }

        private void lblNvHWnd_MouseUp(object sender, MouseEventArgs e)
        {
            if (!lblNvHWnd.Capture) return;

            // キャプチャを終了し，マウスの下にあるウィンドウハンドルを取得
            lblNvHWnd.Capture = false;
            hNvHwnd = WindowFromPoint(Cursor.Position);
            tbxNvHWnd.Text = hNvHwnd.ToString("X8");

            // さらに親に当たるフレームウィンドウのハンドルも取得
            hNvFrame = GetParent(hNvHwnd);
            tbxFraHWnd.Text = (hNvFrame == IntPtr.Zero) ? "(null)" : hNvFrame.ToString("X8");
        }

        //
        // NV.exe が作成する Mutex をチェックするために，OnIdle イベントに
        // Idle時のイベントハンドラを追加し，Mutexをチェック
        //
        private void OnIdle(object s, EventArgs e)
        {
            // Mutex をチェックすることで，NV.exeの状態を調べる
            string strMsg = "";
            IntPtr hMutex = IntPtr.Zero;

            if ((hMutex = OpenMutex(MUTEX_ALL_ACCESS, false, MUTEX_MONITORING)) != IntPtr.Zero)
            {
                CloseHandle(hMutex);
                strMsg = "MONITORING ";
            }
            if ((hMutex = OpenMutex(MUTEX_ALL_ACCESS, false, MUTEX_TERPUMP)) != IntPtr.Zero)
            {
                CloseHandle(hMutex);
                strMsg += "TERPUMP ";
            }
            if ((hMutex = OpenMutex(MUTEX_ALL_ACCESS, false, MUTEX_OPENFILE)) != IntPtr.Zero)
            {
                CloseHandle(hMutex);
                strMsg += "OPENFILE ";
            }
            if ((hMutex = OpenMutex(MUTEX_ALL_ACCESS, false, MUTEX_MANUATTR)) != IntPtr.Zero)
            {
                CloseHandle(hMutex);
                strMsg += "MANUATTR ";
            }
            if ((hMutex = OpenMutex(MUTEX_ALL_ACCESS, false, MUTEX_TOPMOSTWND)) != IntPtr.Zero)
            {
                CloseHandle(hMutex);
                strMsg += "TOPMOSTWND ";
            }
            if ((hMutex = OpenMutex(MUTEX_ALL_ACCESS, false, MUTEX_QUICKDELIVERY)) != IntPtr.Zero)
            {
                CloseHandle(hMutex);
                strMsg += "QUICKDELIVERY ";
            }
            tbxNvStatus.Text = strMsg;
        }

        //
        // ファイルオープンダイアログから，NV.exeを起動する
        //
        private void btnExecPaperChart_Click(object sender, EventArgs e)
        {
            OpenFileDialog ofd = new OpenFileDialog();

            ofd.FileName = "Nv.exe";
            ofd.InitialDirectory = @"D:\paperChart";
            ofd.Filter = "アプリケーション(*.exe)|*.exe|すべてのファイル(*.*)|*.*";
            ofd.FilterIndex = 1;
            ofd.Title = "実行するアプリケーションを選択してください";
            ofd.Multiselect = false;

            //ダイアログボックスを閉じる前に現在のディレクトリを復元するようにする
            ofd.RestoreDirectory = true;
            ofd.CheckFileExists = true;
            ofd.CheckPathExists = true;

            //ダイアログを表示する
            if (ofd.ShowDialog() == DialogResult.OK)
            {
                //選択された実行ファイルをプロセスとして実行
                Process proc = new Process();
                proc.StartInfo.FileName = ofd.FileName;
                proc.StartInfo.Arguments = "";
                proc.StartInfo.UseShellExecute = false;

                proc.Start();
                proc.WaitForInputIdle();     // しばらく待つ(念のため)

                //ウィンドウハンドルが取得できるか，生成したプロセスが終了するまで待つ
                while (proc.MainWindowHandle == IntPtr.Zero && proc.HasExited == false)
                {
                    System.Threading.Thread.Sleep(1);
                    proc.Refresh();
                }

                // 起動したプロセスのウィンドウハンドルは NV.exe の親フレームハンドル
                hNvFrame = proc.MainWindowHandle;
                tbxFraHWnd.Text = hNvFrame.ToString("X8");

                // 親フレームから NV.exeのメイン(View)ウィンドウハンドルを取得する
                // NV.exeは，親フレーム内に 唯一のメインウィンドウを持っている．
                EnumChildWindows(hNvFrame, GetChildHWnd, IntPtr.Zero);
            }
        }

        // EnumChildWindow()から呼ばれるコールバック関数
        // ChildWindowは一つで，それが NV.exeのメインウィンドウハンドル
        int GetChildHWnd(IntPtr hWnd, IntPtr lParam)
        {
            hNvHwnd = hWnd;
            tbxNvHWnd.Text = hWnd.ToString("X8");
            return 0;
        }

        private void btnSetNow_Click(object sender, EventArgs e)
        {
            // 現在時刻をセット
            dtpEventTime.Value = DateTime.Now;
        }

        private void dtpEventTime_ValueChanged(object sender, EventArgs e)
        {
            tbxUnixTime.Text = GetUnixTime(dtpEventTime.Value).ToString();
        }

        //
        // paperChart をコマンドボタンから操作する
        //
        private int SendPpcMenuCommand(Int32 nCmd)
        {
            if (hNvFrame == IntPtr.Zero)
            {
                MessageBox.Show("NV.exe のハンドルを取得して実行してください",
                    "NV.exeハンドル取得エラー", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return 0;
            }
            SendMessage(hNvFrame, WM_COMMAND, (IntPtr)nCmd, IntPtr.Zero);
            return 1;
        }

        //
        //  NVへのデータ送信
        //
        private void SendTextData(string strMsg)
        {
            //
            // paperChart(NV.exe)では 文字列は ShiftJIS のみ受信可能なため， 
            // 送信する文字列を，内部コード(UniCODE)から SJISに変換してバイト配列に格納
            // 注) C# 文字列は C++での文字列終端 '\0'を含んでいない．
            //
            byte[] srcData = Encoding.GetEncoding("Shift_JIS").GetBytes(strMsg);
            int iLen = srcData.Length;                  // '\0' を含まないバイトサイズ

            //
            // アンマネージドメモリーを確保して，バイト配列をコピー
            // SendMessage() を呼び出した後，配列を解放
            //
            IntPtr memPtr = Marshal.AllocHGlobal(iLen + 1);   // 終端'\0'を含むサイズ
            Marshal.Copy(srcData, 0, memPtr, iLen);     // 文字列内容をコピー
            Marshal.WriteByte(memPtr, iLen, 0);         // 終端'\0'の格納

            //
            // COPYDATASTRUCT にアンマネージドメモリー上のデータをセット
            //
            COPYDATASTRUCT cds = new COPYDATASTRUCT();
            cds.dwData = (IntPtr)0x15;                  // 送信コマンドコード
            cds.cbData = (UInt32)iLen + 1;              // 文字列末のNull文字を含むサイズ
            cds.lpData = memPtr;                        // メモリーのポインタを設定

            //
            // NV.exe の ウィンドウハンドルへ WM_COPYDATA メッセージを送信
            //
            SendMessage(hNvHwnd, WM_COPYDATA, (IntPtr)this.Handle, ref cds);

            //
            // 使用したアンマネージドメモリーを解放
            //
            Marshal.FreeHGlobal(memPtr);
        }

        private void btnSendMessage_Click(object sender, EventArgs e)
        {
            if (hNvHwnd == IntPtr.Zero)
            {
                MessageBox.Show("NV.exe のハンドルを取得して実行してください", 
                    "NV.exeハンドル取得エラー", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return;
            }

            string strMsg = tbxMessage.Text;
            if (strMsg == "")
            {
                MessageBox.Show("送信用テキストを設定してください", 
                    "送信エラー", MessageBoxButtons.OK, MessageBoxIcon.Hand);
                return;
            }
            SendTextData(strMsg);
        }

        // サンプルデータが選択されたとき
        private void cbxSelectSample_SelectedIndexChanged(object sender, EventArgs e)
        {
            PpcComboItem item = (PpcComboItem)((ComboBox)sender).SelectedItem;
            string value = item.Value;

            // 時刻を取得
            long lTime = (tbxUnixTime.Text == "") ? GetUnixTime(DateTime.Now) : long.Parse(tbxUnixTime.Text);
            tbxMessage.Text = (value.IndexOf("{0}") > 0) ? string.Format(value, lTime) : value;
        }

        private void btnSendPpcMenu_Click(object sender, EventArgs e)
        {
            PpcMenuItem item = (PpcMenuItem)(cbxPpcMenuItem.SelectedItem);
            if (item == null)
            {
                MessageBox.Show("送信するメニューを選択してください", "メニュー選択", 
                                    MessageBoxButtons.OK, MessageBoxIcon.Asterisk);
                return;
            }
            SendPpcMenuCommand(item.Value);
        }

    }

    // ----- 以下，コンボボックスデータ用クラス -----
    
    //
    // ComboBox 追加用オブジェクト
    //
    public class PpcComboItem
    {
        // プロパティ
        public string Text { get; set; }
        public string Value { get; set; }
        
        // コンストラクタ
        public PpcComboItem(string text, string value)
        {
            this.Text = text;
            this.Value = value;
        }
        // ToStringメソッドのオーバーライド
        public override string ToString()
        {
            return this.Text;
        }
    }

    //
    // ComboBox メニュー用オブジェクト
    //
    public class PpcMenuItem
    {
        // プロパティ
        public string Text { get; set; }
        public Int32 Value { get; set; }

        // コンストラクタ
        public PpcMenuItem(string text, Int32 value)
        {
            this.Text = text;
            this.Value = value;
        }
        // ToStringメソッドのオーバーライド
        public override string ToString()
        {
            return this.Text;
        }
    }
}
